;; Generated package description from multishell.el  -*- no-byte-compile: t -*-
(define-package "multishell" "1.1.9.0.20210802.191646" "Easily use multiple shell buffers, local and remote" '((cl-lib "0.5")) :authors '(("Ken Manheimer" . "ken.manheimer@gmail.com")) :maintainer '("Ken Manheimer" . "ken.manheimer@gmail.com") :keywords '("processes") :url "https://github.com/kenmanheimer/EmacsMultishell")
